package com.aidlab.aidlab_sdk

import android.os.Handler
import android.os.Looper
import com.aidlab.sdk.communication.*
import com.aidlab.sdk.internal.InternalDelegate
import com.aidlab.sdk.internal.InternalUpdateDelegate
import com.aidlab.sdk.communication.DisconnectReason
import io.flutter.Log
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.plugin.common.BinaryMessenger

class AidlabHandler(
    val synchronizationDelegate: AidlabSynchronizationDelegate,
    binaryMessenger: BinaryMessenger
) : AidlabDelegate, InternalDelegate, InternalUpdateDelegate, PlatformAidlab {

    private var released = false

    private var aidlabListener: PlatformAidlabDelegate? = null
    private var internalListener: PlatformAidlabInternalDelegate? = null
    private var updateListener: PlatformAidlabUpdateDelegate? = null

    private var connectedAidlab: IAidlab? = null
    private var isConnected = false

    init {
        released = false
        aidlabListener = PlatformAidlabDelegate(binaryMessenger)
        internalListener = PlatformAidlabInternalDelegate(binaryMessenger)
        updateListener = PlatformAidlabUpdateDelegate(binaryMessenger)
        PlatformAidlab.setUp(binaryMessenger, this)
    }

    fun release() {
        released = true
        connectedAidlab?.disconnect()
        aidlabListener = null
    }

    override fun disconnect() {
        connectedAidlab?.disconnect()
    }

    override fun send(sendRequest: SendRequest) {
        connectedAidlab?.send(sendRequest.command)
    }

    override fun setECGFiltrationMethod(ecgFiltrationRequest: EcgFiltrationRequest) {
        connectedAidlab?.setECGFiltrationMethod(ECGFiltrationMethod.values()[ecgFiltrationRequest.method.toInt()])
    }

    override fun startSynchronization() {
        connectedAidlab?.startSynchronization()
    }

    override fun stopSynchronization() {
        connectedAidlab?.stopSynchronization()
    }

    override fun requestForFirmwareInfo(arg: FirmwareInfoRequest) {
        connectedAidlab?.internalAidlab?.requestForFirmwareInfo(arg.url)
    }

    override fun startFirmwareUpdate(arg: FirmwareUpdateRequest) {
        connectedAidlab?.internalAidlab?.startFirmwareUpdate(arg!!.url, arg.updatableVersion)
    }

    override fun didConnectAidlab(aidlab: IAidlab) {
        if (released) {
            aidlab.disconnect()
            return
        }
        isConnected = true
        connectedAidlab = aidlab
        aidlab.internalAidlab.internalDelegate = this
        aidlab.internalAidlab.updateDelegate = this

        connectedAidlab?.aidlabSynchronizationDelegate = synchronizationDelegate
        val connectEventArg = ConnectEvent(aidlab.firmwareRevision!!, aidlab.hardwareRevision!!, aidlab.serialNumber!!, aidlab.address())
        aidlabListener?.didConnect(connectEventArg) {}
    }

    override fun didDisconnectAidlab(aidlab: IAidlab, reason: DisconnectReason) {
        isConnected = false

        connectedAidlab?.aidlabSynchronizationDelegate = null
        connectedAidlab = null
        aidlabListener?.didDisconnect(DisconnectEvent(reason.value.toLong())) {}
    }

    override fun didReceiveECG(aidlab: IAidlab, timestamp: Long, values: FloatArray) {
        aidlabListener?.didReceiveEcg(EcgEvent(timestamp, values.map { it.toDouble() }.toDoubleArray())) {}
    }

    override fun didDetectExercise(aidlab: IAidlab, exercise: Exercise) {
        aidlabListener?.didDetectExercise(ExerciseEvent(exercise.value.toLong())) {}
    }

    override fun didReceiveAccelerometer(
        aidlab: IAidlab,
        timestamp: Long,
        ax: Float,
        ay: Float,
        az: Float
    ) {
        aidlabListener?.didReceiveAccelerometer(Utils.createVectorEvent(timestamp, ax, ay, az)) {}
    }

    override fun didReceiveActivity(aidlab: IAidlab, timestamp: Long, activity: ActivityType) {
        aidlabListener?.didReceiveActivity(ActivityEvent(timestamp, activity.value.toLong())) {}
    }

    override fun didReceiveBatteryLevel(aidlab: IAidlab, stateOfCharge: Int) {
        aidlabListener?.didReceiveBatteryLevel(BatteryLevelEvent(stateOfCharge.toLong())) {}
    }

    override fun didReceiveBodyPosition(
        aidlab: IAidlab,
        timestamp: Long,
        bodyPosition: BodyPosition
    ) {
        aidlabListener?.didReceiveBodyPosition(BodyPositionEvent(timestamp, bodyPosition.value.toLong())) {}
    }

    override fun didReceiveCommand(aidlab: IAidlab) {
        Log.d("AidlabHandler", "didReceiveCommand")
        aidlabListener?.didReceiveCommand(EmptyEvent(0)) {}
    }

    override fun didReceiveError(error: String) {
        aidlabListener?.didReceiveError(ErrorEvent(error)) {}
    }

    override fun didReceiveFirmwareInformation(updateAvailable: Boolean, updatableVersion: String) {
        Handler(Looper.getMainLooper()).post {
            internalListener?.didReceiveFirmwareInformation(FirmwareInformationEvent(updateAvailable, updatableVersion)) {}
        }
    }

    override fun didReceiveGyroscope(
        aidlab: IAidlab,
        timestamp: Long,
        qx: Float,
        qy: Float,
        qz: Float
    ) {
        aidlabListener?.didReceiveGyroscope(Utils.createVectorEvent(timestamp, qx, qy, qz)) {}
    }

    override fun didReceiveHeartRate(aidlab: IAidlab, timestamp: Long, heartRate: Int) {
        aidlabListener?.didReceiveHeartRate(HeartRateEvent(timestamp, heartRate.toLong())) {}
    }

    override fun didReceiveRr(aidlab: IAidlab, timestamp: Long, rr: Int) {
        aidlabListener?.didReceiveRr(RrEvent(timestamp, rr.toLong())) {}
    }

    override fun didReceiveMagnetometer(
        aidlab: IAidlab,
        timestamp: Long,
        mx: Float,
        my: Float,
        mz: Float
    ) {
        aidlabListener?.didReceiveMagnetometer(Utils.createVectorEvent(timestamp, mx, my, mz)) {}
    }

    override fun didReceiveMessage(aidlab: IAidlab, process: String, message: String) {
        aidlabListener?.didReceiveMessage(MessageEvent(process, message)) {}
    }

    override fun didReceiveOrientation(
        aidlab: IAidlab, timestamp: Long,
        roll: Float, pitch: Float, yaw: Float
    ) {
        aidlabListener?.didReceiveOrientation(OrientationEvent(timestamp, roll.toDouble(), pitch.toDouble(), yaw.toDouble())) {}
    }

    override fun didReceivePressure(aidlab: IAidlab, timestamp: Long, values: IntArray) {
        internalListener?.didReceivePressure(PressureEvent(timestamp, values)) {}
    }

    override fun didReceiveQuaternion(
        aidlab: IAidlab,
        timestamp: Long,
        qw: Float,
        qx: Float,
        qy: Float,
        qz: Float
    ) {
        aidlabListener?.didReceiveQuaternion(QuaternionEvent(timestamp, qw.toDouble(), qx.toDouble(), qy.toDouble(), qz.toDouble())) {}
    }

    override fun didReceiveRespiration(aidlab: IAidlab, timestamp: Long, values: FloatArray) {
        aidlabListener?.didReceiveRespiration(RespirationEvent(timestamp, values.map { it.toDouble() }.toDoubleArray())) {}
    }

    override fun didReceiveRespirationRate(aidlab: IAidlab, timestamp: Long, value: Int) {
        aidlabListener?.didReceiveRespirationRate(RespirationRateEvent(timestamp, value.toLong())) {}
    }

    override fun didReceiveSignalQuality(aidlab: IAidlab, timestamp: Long, value: Int) {
        aidlabListener?.didReceiveSignalQuality(SignalQualityEvent(timestamp, value.toLong())) {}
    }

    override fun didReceiveSkinTemperature(aidlab: IAidlab, timestamp: Long, value: Float) {
        aidlabListener?.didReceiveSkinTemperature(SkinTemperatureEvent(timestamp, value.toDouble())) {}
    }

    override fun didReceiveSoundVolume(aidlab: IAidlab, timestamp: Long, value: Int) {
        aidlabListener?.didReceiveSoundVolume(SoundVolumeEvent(timestamp, value.toLong())) {}
    }

    override fun didReceiveSteps(aidlab: IAidlab, timestamp: Long, value: Long) {
        aidlabListener?.didReceiveSteps(StepsEvent(timestamp, value)) {}
    }

    override fun pressureWearStateDidChange(aidlab: IAidlab, state: WearState) {
        internalListener?.pressureWearStateDidChange(PressureWearStateEvent(state.value.toLong())) {}
    }

    override fun wearStateDidChange(aidlab: IAidlab, wearState: WearState) {
        aidlabListener?.wearStateDidChange(WearStateEvent(wearState.value.toLong())) {}
    }

    override fun didReceiveSoundFeatures(aidlab: IAidlab, values: FloatArray): Unit {
        val values_ = values.map { it.toDouble() }.toDoubleArray()
        aidlabListener?.didReceiveSoundFeatures(SoundFeaturesEvent(values_)) {}
    }

    override fun didDetectUserEvent(timestamp: Long) {
        aidlabListener?.didDetectUserEvent(UserEvent(timestamp)) {}
    }

    override fun didFinishUpdate() {
        updateListener?.didFinishUpdate(EmptyEvent(0)) {}
    }

    override fun didReceiveUpdateError(error: UpdateError) {
        // Change to the UI thread. Other callback are on the UI thread by default
        Handler(Looper.getMainLooper()).post {
            updateListener?.didReceiveUpdateError(UpdateEvent(error.value.toLong())) {}
        }
    }

    override fun didReceiveUpdateProgress(progress: Int) {
        updateListener?.didReceiveUpdateProgress(UpdateProgress(progress.toLong())) {}
    }
}