//
//  AidlabHandler.swift
//  AidlabSDK
//
//  Created by Szymon Gęsicki on 05/02/2021.
//  Copyright © 2020 Aidlab. All rights reserved.
//

import Flutter
import AidlabSDK

public class AidlabHandler: AidlabDelegate, PlatformAidlab, InternalDelegate, InternalUpdateDelegate {
    
    private var synchronizationDelegate: AidlabSynchronizationDelegate
    private var released: Bool = false
    private var aidlabListener: PlatformAidlabDelegate?
    private var internalListener: PlatformAidlabInternalDelegate?
    private var updateListener: PlatformAidlabUpdateDelegate?
    private var connectedAidlab: IAidlab?
    private var isConnected: Bool = false
    
    init(synchronizationDelegate: AidlabSynchronizationDelegate) {
        self.synchronizationDelegate = synchronizationDelegate
    }
    
    func initial(binding: FlutterBinaryMessenger) {
        released = false
        aidlabListener = PlatformAidlabDelegate(binaryMessenger: binding)
        internalListener = PlatformAidlabInternalDelegate(binaryMessenger: binding)
        updateListener = PlatformAidlabUpdateDelegate(binaryMessenger: binding)
        PlatformAidlabSetup.setUp(binaryMessenger: binding, api: self)
    }
    
    func release() {
        released = true
        connectedAidlab?.disconnect()
        aidlabListener = nil
    }
    
    func disconnect() throws {
        connectedAidlab?.disconnect()
    }
    
    func send(sendRequest: SendRequest) throws {
        connectedAidlab?.send(sendRequest.command)
    }
    
    func setECGFiltrationMethod(ecgFiltrationRequest: EcgFiltrationRequest) throws {
        let ecgFiltrationMethod: ECGFiltrationMethod = (ecgFiltrationRequest.method == 0) ? .normal : .aggressive

        connectedAidlab?.setECGFiltrationMethod(ecgFiltrationMethod)
    }
    
    func startSynchronization() throws {
        connectedAidlab?.startSynchronization()
    }
    
    func stopSynchronization() throws {
        connectedAidlab?.stopSynchronization()
    }
    
    func requestForFirmwareInfo(firmwareInfoRequest: FirmwareInfoRequest) throws {
        connectedAidlab?.internalAidlab.requestForFirmwareInfo(url: firmwareInfoRequest.url)
    }
    
    func startFirmwareUpdate(firmwareUpdateRequest: FirmwareUpdateRequest) throws {
        connectedAidlab?.internalAidlab.startFirmwareUpdate(url: firmwareUpdateRequest.url, updatableVersion: firmwareUpdateRequest.updatableVersion)
    }

    public func didConnect(_ aidlab: IAidlab) {

        if released {
            aidlab.disconnect()
            return
        }
        
        isConnected = true
        connectedAidlab = aidlab
        
        connectedAidlab?.internalAidlab.internalDelegate = self
        connectedAidlab?.internalAidlab.updateInternalDelegate = self
        connectedAidlab?.aidlabSynchronizationDelegate = synchronizationDelegate

        aidlabListener?.didConnect(connectEvent: ConnectEvent(firmwareRevision: aidlab.firmwareRevision!, hardwareRevision: aidlab.hardwareRevision!, serialNumber: aidlab.serialNumber!, address: aidlab.address!.uuidString), completion: { _ in })
    }
    
    public func didDisconnect(_ aidlab: IAidlab, reason: DisconnectReason) {
        isConnected = false
        connectedAidlab?.aidlabSynchronizationDelegate = nil
        connectedAidlab = nil
        aidlabListener?.didDisconnect(voidEvent: DisconnectEvent(disconnectReason: Int64(reason.rawValue)), completion: { _ in })
    }
    
    public func didReceiveECG(_ aidlab: IAidlab, timestamp: UInt64, values: [Float]) {
        let data = Data(values.withUnsafeBufferPointer { Data(buffer: $0) })
        aidlabListener?.didReceiveEcg(ecgEvent: EcgEvent(timestamp: Int64(timestamp), values: FlutterStandardTypedData(float32: data)), completion: { _ in })
    }
    
    public func didReceiveRespiration(_ aidlab: IAidlab, timestamp: UInt64, values: [Float]) {

        let data = Data(values.withUnsafeBufferPointer { Data(buffer: $0) })
        aidlabListener?.didReceiveRespiration(respirationEvent: RespirationEvent(timestamp: Int64(timestamp), values: FlutterStandardTypedData(float32: data)), completion: { _ in })
    }
    
    public func didReceiveBatteryLevel(_ aidlab: IAidlab, stateOfCharge: UInt8) {
        aidlabListener?.didReceiveBatteryLevel(batteryLevelEvent: BatteryLevelEvent(stateOfCharge: Int64(stateOfCharge)), completion: { _ in })
    }
    
    public func didReceiveSteps(_ aidlab: IAidlab, timestamp: UInt64, value: UInt64) {
        aidlabListener?.didReceiveSteps(stepsEvent: StepsEvent(timestamp: Int64(timestamp), value: Int64(value)), completion: { _ in })
    }
    
    public func didReceiveSkinTemperature(_ aidlab: IAidlab, timestamp: UInt64, value: Float) {
        aidlabListener?.didReceiveSkinTemperature(skinTemperatureEvent: SkinTemperatureEvent(timestamp: Int64(timestamp), value: Double(value)), completion: { _ in })
    }
    
    public func didReceiveAccelerometer(_ aidlab: IAidlab, timestamp: UInt64, ax: Float, ay: Float, az: Float) {
        aidlabListener?.didReceiveAccelerometer(vectorEvent: VectorEvent(timestamp: Int64(timestamp), x: Double(ax), y: Double(ay), z: Double(az)), completion: { _ in })
    }
    
    public func didReceiveGyroscope(_ aidlab: IAidlab, timestamp: UInt64, qx: Float, qy: Float, qz: Float) {
        aidlabListener?.didReceiveGyroscope(vectorEvent: VectorEvent(timestamp: Int64(timestamp), x: Double(qx), y: Double(qy), z: Double(qz)), completion: { _ in })
    }
    
    public func didReceiveMagnetometer(_ aidlab: IAidlab, timestamp: UInt64, mx: Float, my: Float, mz: Float) {
        aidlabListener?.didReceiveMagnetometer(vectorEvent: VectorEvent(timestamp: Int64(timestamp), x: Double(mx), y: Double(my), z: Double(mz)), completion: { _ in })
    }
    
    public func didReceiveQuaternion(_ aidlab: IAidlab, timestamp: UInt64, qw: Float, qx: Float, qy: Float, qz: Float) {
        aidlabListener?.didReceiveQuaternion(quaternionEvent: QuaternionEvent(timestamp: Int64(timestamp), qw: Double(qw), qx: Double(qx), qy: Double(qy), qz: Double(qz)), completion: { _ in })
    }
    
    public func didReceiveOrientation(_ aidlab: IAidlab, timestamp: UInt64, roll: Float, pitch: Float, yaw: Float) {
        aidlabListener?.didReceiveOrientation(orientationEvent: OrientationEvent(timestamp: Int64(timestamp), roll: Double(roll), pitch: Double(pitch), yaw: Double(yaw)), completion: { _ in })
    }
    
    public func didReceiveBodyPosition(_ aidlab: IAidlab, timestamp: UInt64, bodyPosition: BodyPosition) {
        aidlabListener?.didReceiveBodyPosition(bodyPositionEvent: BodyPositionEvent(timestamp: Int64(timestamp), bodyPosition: Int64(bodyPosition.rawValue)), completion: { _ in })
    }
    
    public func didReceiveHeartRate(_ aidlab: IAidlab, timestamp: UInt64, heartRate: Int32) {
        aidlabListener?.didReceiveHeartRate(heartRateEvent: HeartRateEvent(timestamp: Int64(timestamp), heartRate: Int64(heartRate)), completion: { _ in })
    }
    
    public func didReceiveRr(_ aidlab: IAidlab, timestamp: UInt64, rr: Int32) {
        aidlabListener?.didReceiveRr(rrEvent: RrEvent(timestamp: Int64(timestamp), rr: Int64(rr)), completion: { _ in })
    }
    
    public func didReceiveRespirationRate(_ aidlab: IAidlab, timestamp: UInt64, value: UInt32) {
        aidlabListener?.didReceiveRespirationRate(respirationRateEvent: RespirationRateEvent(timestamp: Int64(timestamp), value: Int64(value)), completion: { _ in })
    }
    
    public func didReceiveSoundVolume(_ aidlab: IAidlab, timestamp: UInt64, soundVolume: UInt16) {
        aidlabListener?.didReceiveSoundVolume(soundVolumeEvent: SoundVolumeEvent(timestamp: Int64(timestamp), soundVolume: Int64(soundVolume)), completion: { _ in })
    }
    
    public func didDetectUserEvent(_ aidlab: IAidlab, timestamp: UInt64) {
        aidlabListener?.didDetectUserEvent(userEvent: UserEvent(timestamp: Int64(timestamp)), completion: { _ in })
    }
    
    public func didReceivePressure(_ aidlab: IAidlab, timestamp: UInt64, values: [Int32]) {
        let data = Data(values.withUnsafeBufferPointer { Data(buffer: $0) })
        internalListener?.didReceivePressure(pressureEvent: PressureEvent(timestamp: Int64(timestamp), values: FlutterStandardTypedData(int32: data)), completion: { _ in })
    }
    
    public func didDetect(_ aidlab: IAidlab, exercise: Exercise) {
        aidlabListener?.didDetectExercise(exerciseEvent: ExerciseEvent(exercise: Int64(exercise.rawValue)), completion: { _ in })
    }
    
    public func didDetect(_ aidlab: IAidlab, timestamp: UInt64, activity: ActivityType) {
        aidlabListener?.didReceiveActivity(activityEvent: ActivityEvent(timestamp: Int64(timestamp), activityType: Int64(activity.rawValue)), completion: { _ in })
    }
    
    public func didReceiveError(_ aidlab: IAidlab, error: Error) {
        aidlabListener?.didReceiveError(errorEvent: ErrorEvent(error: error.localizedDescription), completion: { _ in })
    }
    
    public func didReceiveRSSI(_ aidlab: IAidlab, rssi: Int32) {
        /// TODO !
    }
    
    public func wearStateDidChange(_ aidlab: IAidlab, state: WearState) {
        aidlabListener?.wearStateDidChange(wearStateEvent: WearStateEvent(wearState: Int64(state.rawValue)), completion: { _ in })
    }
    
    public func pressureWearStateDidChange(_ aidlab: IAidlab, state: WearState) {
        internalListener?.pressureWearStateDidChange(wearStateEvent: PressureWearStateEvent(wearState: Int64(state.rawValue)), completion: { _ in })
    }
    
    public func didReceiveCommand(_ aidlab: IAidlab) {
        aidlabListener?.didReceiveCommand(voidEvent: EmptyEvent(nothing: 0), completion: { _ in })
    }
    
    public func didReceiveMessage(_ aidlab: IAidlab, process: String, message: String) {
        aidlabListener?.didReceiveMessage(messageEvent: MessageEvent(process: process, message: message), completion: { _ in })
    }
    
    public func didReceiveFirmwareInformation(updateAvailable: Bool, updatableVersion: String) {
        internalListener?.didReceiveFirmwareInformation(firmwareInformationEvent: FirmwareInformationEvent(updateAvailable: Bool(updateAvailable), updatableVersion: updatableVersion), completion: { _ in })
    }
    
    public func didReceiveSoundFeatures(_ aidlab: IAidlab, values: [Float]) {
        let data = Data(values.withUnsafeBufferPointer { Data(buffer: $0) })
        aidlabListener?.didReceiveSoundFeatures(soundFeaturesEvent: SoundFeaturesEvent(values: FlutterStandardTypedData(float32: data)), completion: { _ in })
    }
    
    public func didReceiveSignalQuality(_ timestamp: UInt64, value: Int32) {
        aidlabListener?.didReceiveSignalQuality(signalQualityEvent: SignalQualityEvent(timestamp: Int64(timestamp), value: Int64(value)), completion: { _ in })
    }
    
    public func didReceiveUpdateError(error: UpdateError) {
        updateListener?.didReceiveUpdateError(event: UpdateEvent(error: Int64(error.rawValue)), completion: { _ in })
    }
    
    public func didReceiveUpdateProgress(progress: UInt8) {
        updateListener?.didReceiveUpdateProgress(progress: UpdateProgress(progress: Int64(progress)), completion: { _ in })
    }
    
    public func didFinishUpdate() {
        updateListener?.didFinishUpdate(nothing: EmptyEvent(nothing: 0), completion: { _ in })
    }
}
